# coding:utf-8
'''
Created on 14.04.2011

@author: kirov
'''
import os
import uuid
import csv
import cStringIO
import codecs

from django.conf import settings
from django.db.models import Q, Count, Avg, Max, Min, Sum

try:
    import xlwt
except ImportError:
    class FakeXlwt(object):
        def __getattr__(self, *args):
            raise RuntimeError(u'xlwt is not installed!')
    xlwt = FakeXlwt()


class UnicodeWriter:

    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([unicode(s).encode("utf-8") for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


class RecordProxy(object):

    '''
    Прокси объект записи отображающей группировку
    '''
    def __init__(self, *args, **kwargs):
        self.id = 0  # ID записи в базе
        self.index = 0  # индекс записи в раскрытом дереве (заполняется при выводе)
        self.lindex = 0  # индекс записи на уровене (заполняется при выводе)
        self.indent = 0  # уровень (заполняется при выводе)
        self.is_leaf = False  # признак конечной записи (больше не раскрывается) (заполняется при выводе)
        self.expanded = False  # признак что элемент развернут (заполняется при выводе)
        self.count = 0  # количество дочерних узлов, если они есть
        self.grouped = []
            # Список атрибутов, по которым происходит группировка
        self.init_component(*args, **kwargs)

    def init_component(self, *args, **kwargs):
        '''
        Заполняет атрибуты экземпляра значениями в kwargs,
        если они проинициализированы ранее
        '''
        for k, v in kwargs.items():
            assert k in dir(self) and not callable(getattr(self, k)), \
                'Instance attribute "%s" should be defined in class "%s"!' \
                % (k, self.__class__.__name__)
            self.__setattr__(k, v)

    def load(self, data):
        '''
        Загрузка записи в объект
        '''
        pass

    def calc(self):
        '''
        Пост-обработка записи, когда все реквизиты заполнены
        '''
        pass


class DictToRecordProxy(RecordProxy):
    """
    Прокси объект для данных, переданных в виде словаря
    """
    def load(self, data):
        """
        Копирует все записи словаря в атрибуты объекта
        """
        self.__dict__.update(data)


class GroupingRecordProvider(object):

    '''
    Базовый класс провайдера данных
    '''
    proxy_class = RecordProxy
    data_source = None
    count_totals = False
    aggregates = {}
    xls_style = {}

    def __init__(
            self, proxy=None, data=None, totals=None,
            aggregates=None, detail_attrs_map=None):
        if proxy:
            self.proxy_class = proxy
        if data is not None:
            self.data_source = data
        if totals is not None:
            self.count_totals = totals
        if aggregates is not None:
            self.aggregates = aggregates

        # Словарь, определяющий дополнительные поля,
        # которые добавляются в узловые записи,
        # создаваемые самим провайдером.
        # Они могут пригодиться для последующей обработки узлов.
        # Например, detail_attrs_map={'a_name': 'a_id', 'b_name': 'b_id'}
        # Если группировка будет по полям a_name и b_name,
        # то вместе с ними будет добавлены a_id и b_id.
        self.detail_attrs_map = detail_attrs_map or {}

    def get_data(self, *args, **kwargs):
        return self.data_source

    def getattr(self, obj, attr):
        """
        Получить атрибут объекта или элемент словаря
        """
        if isinstance(obj, dict):
            if attr in obj:
                return obj[attr]
            else:
                return None
        else:
            if hasattr(obj, attr):
                return getattr(obj, attr)
            else:
                return None

    def setattr(self, obj, attr, value):
        """
        Установить атрибут объекта или элемент словаря
        """
        if isinstance(obj, dict):
            obj[attr] = value
        else:
            setattr(obj, attr, value)

    def calc(self, obj):
        """
        Вычислить остальные атрибуты объекта или словаря
        """
        if isinstance(obj, dict):
            pass
        else:
            if hasattr(obj, 'calc') and callable(obj.calc):
                obj.calc()

    def load(self, destination, source):
        """
        Загрузка объекта destination из атрибутов объекта source
        """
        EXCLUDED = ('index', 'lindex', 'indent',
                    'is_leaf', 'expanded', 'count')
        if isinstance(destination, dict):
            for attr in destination:
                if not attr in EXCLUDED:
                    if isinstance(source, dict):
                        if attr in source:
                            destination[attr] = source[attr]
                    else:
                        if hasattr(source, attr):
                            destination[attr] = getattr(source, attr)
        else:
            if hasattr(destination, 'load') and callable(destination.load):
                destination.load(source)

    def create_record(self, *args, **kwargs):
        if isinstance(self.proxy_class, dict):
            rec = self.proxy_class.copy()
            rec['index'] = 0
            rec['lindex'] = 0
            rec['indent'] = 0
            rec['is_leaf'] = False
            rec['expanded'] = False
            rec['count'] = 0
            return rec
        else:
            return self.proxy_class(*args, **kwargs)

    def reader(self, grouped, offset, level_index, level_keys, begin, end, aggregates, sorting):
        pass

    def counter(self, grouped, level_index, level_keys, expandedItems):
        pass

    def indexer(self, grouped, level_index, level_keys, expandedItems, aggregates, sorting):
        pass

    def get_elements(self, begin, end, grouped, expanded, sorting):
        '''
        Основной метод получения данных
        '''
        return self.__get_elements(grouped, begin, 0, {'count': -1, 'id': -1, 'expandedItems': expanded}, begin, end, [], self.aggregates, sorting)

    def __get_elements(self, grouped, offset, level_index, level, begin, end, keys, aggregates, sorting):
        res = []
        all_out = False  # признак того, что все необходимые элементы выведены, идет подсчет общего количества открытых элементов
        # print 'get_elements(): grouped=%s, offset=%s, level_index=%s,
        # level=%s, begin=%s, end=%s, keys=%s' % (grouped, offset, level_index,
        # level, begin, end, keys)

        # необходимо узнавать индекс элемента с кодом exp['id'] в текущем уровне, при текущей сортировке и располагать элементы соответственно
        if 'items' not in level:
            level['items'] = self.indexer(grouped, level_index, keys + [level[
                                          'id']] if level['id'] != -1 else [], level['expandedItems'], aggregates, sorting)
            removeExpandedItems = []
            for exp in level['expandedItems']:
                if exp['id'] in level['items']:
                    exp['index'] = level['items'].index(exp['id'])
                else:
                    # развернутый элемент отсутствует в уровне (видимо фильтр сработал, или что-то еще) - что делать?!
                    removeExpandedItems.append(exp)

            for exp in removeExpandedItems:
                level['expandedItems'].remove(exp)

            # теперь выстроим в порядке индексов
            exp_sort = sorted(level['expandedItems'], key=lambda x: x['index'])
            level['expandedItems'] = exp_sort

        # пройдемся по развернутым элементам
        i = 0
        while i < len(level['expandedItems']):
            exp = level['expandedItems'][i]

            # на текущий момент необходимо вычислить количество дочерних элементов
            if exp['count'] == -1:
                exp['count'] = self.counter(grouped, level_index + 1, (keys + [level[
                                            'id']] if level['id'] != -1 else []) + [exp['id']], exp['expandedItems'])

            if all_out or not exp['count']:
                i = i + 1
                continue

            # 1) может диапазон уже пройден
            if end <= exp['index']:
                # выдать диапазон с begin по end
                # print '1) диапазон уже пройден'
                # print 'offset=%s, begin=%s, end=%s, exp=%s, keys=%s' %
                # (offset, begin, end, exp, keys)
                list = self.reader(grouped, offset + len(res) - begin, level_index, keys + [
                                   level['id']] if level['id'] != -1 else [], begin, end, aggregates, sorting)
                # если выдали раскрытый элемент, то установим у него признак раскрытости
                if end == exp['index']:
                    self.setattr(list[-1], 'expanded', True)
                res.extend(list)
                # переходим к след. развернутому элементу
                i = i + 1
                all_out = True
                continue

            # 2) может интервал переходит с предыдущего
            if begin <= exp['index'] and end > exp['index']:
                # print '2) интервал переходит с предыдущего'
                # print 'offset=%s, begin=%s, end=%s, exp=%s, keys=%s' % (offset, begin, end, exp, keys)
                # выдадим известный диапазон, а остальное продолжим искать
                list = self.reader(grouped, offset + len(res) - begin, level_index, keys + [level[
                                   'id']] if level['id'] != -1 else [], begin, exp['index'], aggregates, sorting)
                # если выдали раскрытый элемент, то установим у него признак раскрытости
                self.setattr(list[-1], 'expanded', True)
                res.extend(list)
                begin = exp['index'] + 1
                continue

            # 3) попадаем ли мы в раскрытый уровень
            if begin >= exp['index'] and end <= exp['index'] + exp['count']:
                # print '3) мы попадаем в раскрытый уровень'
                # print 'offset=%s, begin=%s, end=%s, exp=%s, keys=%s' % (offset, begin, end, exp, keys)
                # переходим искать на след. уровень
                list, total = self.__get_elements(grouped, offset + len(res), level_index + 1, exp, begin - exp[
                                                  'index'] - 1, end - exp['index'] - 1, keys + [level['id']] if level['id'] != -1 else [], aggregates, sorting)
                res.extend(list)
                # переходим к след. развернутому элементу
                i = i + 1
                all_out = True
                continue

            # 4) если частично попадаем в раскрытый
            if begin <= exp['index'] + exp['count'] and end > exp['index'] + exp['count']:
                # print '4) частично попадаем в раскрытый'
                # print 'offset=%s, begin=%s, end=%s, exp=%s, keys=%s' % (offset, begin, end, exp, keys)
                # часть переведем искать на след. уровень, остальное продолжим
                list, total = self.__get_elements(grouped, offset + len(res), level_index + 1, exp, begin - exp[
                                                  'index'] - 1, exp['count'] - 1, keys + [level['id']] if level['id'] != -1 else [], aggregates, sorting)
                res.extend(list)
                delta = end - begin - len(list)
                begin = exp[
                    'index'] + 1  # поднимем нижнюю границу до следующего после текущего элемента
                end = begin + \
                    delta  # end - len(list) # сократим верхнюю границу на количество выданных элементов
                i = i + 1
                continue

            # 6) если еще не дошли
            if begin > exp['index'] + exp['count']:
                # print '6) если еще не дошли'
                # print 'offset=%s, begin=%s, end=%s, exp=%s, keys=%s' %
                # (offset, begin, end, exp, keys)
                begin = begin - exp['count']
                end = end - exp['count']

            # переходим к след. развернутому элементу
            i = i + 1

        if level['count'] == -1:
            level['count'] = self.counter(
                grouped, level_index, keys, level['expandedItems'])

        # 5) выдадим из уровеня всё что осталось
        if not all_out and begin <= end and begin < level['count']:
            # print '5) выдадим из уровеня всё что осталось'
            # print 'begin=%s, end=%s, level[count]=%s, len(res)=%s, offset=%s,
            # keys=%s' % (begin,end,level['count'], len(res), offset, keys)
            if end > level['count'] - 1:
                end = level['count'] - 1
            list = self.reader(grouped, offset + len(res) - begin, level_index, keys + [
                               level['id']] if level['id'] != -1 else [], begin, end, aggregates, sorting)
            res.extend(list)

        # можно уже не считать total выше
        total_len = level['count']

        # print 'get_elements()= total=%s, res_count=%s' % (total_len,
        # len(res))

        # если это самый первый уровень и нужно считать общие итоги, то посчитаем выдадим
        if level_index == 0 and self.count_totals:
            totals = self.reader(grouped, 0, -1, [], 0, 1, aggregates, sorting)
            return (res, (total_len, totals))
        else:
            return (res, total_len)

    def get_export_group_text(self, item, grouped_col_name):
        '''
        Форматирование колонки группировки
        '''
        indent_str = "   " * self.getattr(item, 'indent')
        value = self.getattr(item, 'id')
        return "%s %s: %s (%s)" % (indent_str, grouped_col_name, value, self.getattr(item, 'count'))

    EXPORT_XLS = 'xls'
    EXPORT_CSV = 'csv'

    def export_to_file(self, title, columns, total, grouped, expanded, sorting, export_type=EXPORT_XLS):
        if export_type == self.EXPORT_XLS:
            return self.export_to_xls(title, columns, total, grouped, expanded, sorting)
        elif export_type == self.EXPORT_CSV:
            return self.export_to_csv(title, columns, total, grouped, expanded, sorting)
        else:
            return None

    def _tune_xls_worksheet(self, worksheet):
        """
        Здесь можно задать нужные параметры страницы.
        """
        return worksheet

    def export_to_xls (self, title, columns, total, grouped, expanded, sorting):
        """
        Выгрузка таблицы в xls-файл
        """
        w = xlwt.Workbook()
        ws = w.add_sheet('grid')
        ws = self._tune_xls_worksheet(ws)

        title_style = xlwt.easyxf(self.xls_style.get(
            'title') or "font: bold on, height 400;")
        header_style = xlwt.easyxf(self.xls_style.get('header') or
                                   "font: bold on, color-index white;"
                                   "borders: left thick, right thick, top thick, bottom thick;"
                                   "pattern: pattern solid, fore_colour gray80;"
                                   )
        data_style = xlwt.easyxf(self.xls_style.get('data') or
                                 "borders: left thin, right thin, top thin, bottom thin;"
                                 )

        col_count = 0
        total_width = 0
        for column in columns:
            if not column.get("hidden"):
                col_count += 1
                total_width += column.get("width")

        page_width = 30000
        ws.write_merge(0, 0, 0, col_count - 1, title, title_style)
        ws.row(0).height = 500
        columns_cash = []
        columns_title = {}
        index = 0
        for column in columns:
            if not column.get("hidden"):
                ws.write(1, index, column.get('header'), header_style)
                ws.col(index).width = 1.0 * column.get(
                    "width") * page_width / total_width
                columns_title[column["data_index"]] = column.get('header')
                columns_cash.append(column["data_index"])
                index += 1
        # запросим все данные
        data, total = self.get_elements(0, total, grouped, expanded, sorting)
        # вывод данных
        for item in data:
            for idx, k in enumerate(columns_cash):
                # значит это группировочная колонка
                if k == "grouping":
                    if self.getattr(item, 'is_leaf'):
                        v = ""
                    else:
                        col = grouped[self.getattr(item, 'indent')]
                        col_name = columns_title[col]
                        v = self.get_export_group_text(item, col_name)
                    ws.write(self.getattr(
                        item, 'index') + 2, idx, v, data_style)
                else:
                    v = self.getattr(item, k)
                    ws.write(self.getattr(
                        item, 'index') + 2, idx, v, data_style)
        # вывод итогов
        if not isinstance(total, (int, long)):
            total_row = total[1]
            for idx, k in enumerate(columns_cash):
                if k == "grouping":
                    v = ""
                else:
                    v = self.getattr(total_row, k)
                ws.write(total[0] + 2, idx, v, header_style)

        base_name = str(uuid.uuid4())[0:16] + '.xls'

        xls_file_abs = os.path.join(settings.MEDIA_ROOT, base_name)
        w.save(xls_file_abs)
        join = lambda pieces: u'/'.join(s.strip('/') for s in pieces)
        url = join([settings.MEDIA_URL, base_name])
        return url

    def export_to_csv(self, title, columns, total, grouped, expanded, sorting):
        '''
        выгрузка таблицы в csv-файл
        '''
        base_name = str(uuid.uuid4())[0:16] + '.csv'
        file_abs = os.path.join(settings.MEDIA_ROOT, base_name)
        # ws = UnicodeWriter(open(file_abs, 'wb'), delimiter='|',
        # quotechar='"', quoting=csv.QUOTE_MINIMAL)
        ws = UnicodeWriter(open(
            file_abs, 'wb'), delimiter=';', encoding="cp1251")
        col_count = 0
        for column in columns:
            if not column.get("hidden"):
                col_count += 1

        ws.writerow([title])
        columns_cash = []
        columns_title = {}
        index = 0
        row = []
        for column in columns:
            if not column.get("hidden"):
                row.append(column.get('header'))
                columns_title[column["data_index"]] = column.get('header')
                columns_cash.append(column["data_index"])
                index += 1
        ws.writerow(row)
        # запросим все данные
        data, total = self.get_elements(0, total, grouped, expanded, sorting)
        # вывод данных
        for item in data:
            row = []
            for idx, k in enumerate(columns_cash):
                # значит это группировочная колонка
                if k == "grouping":
                    if self.getattr(item, 'is_leaf'):
                        v = ""
                    else:
                        col = grouped[self.getattr(item, 'indent')]
                        col_name = columns_title[col]
                        v = self.get_export_group_text(item, col_name)
                    row.append(v if v else '')
                else:
                    v = self.getattr(item, k)
                    row.append(v if v else '')
            ws.writerow(row)
        # вывод итогов
        if not isinstance(total, (int, long)):
            row = []
            total_row = total[1]
            for idx, k in enumerate(columns_cash):
                if k == "grouping":
                    v = ""
                else:
                    v = self.getattr(total_row, k)
                row.append(v if v else '')
            ws.writerow(row)

        join = lambda pieces: u'/'.join(s.strip('/') for s in pieces)
        url = join([settings.MEDIA_URL, base_name])
        return url


class GroupingRecordModelProvider(GroupingRecordProvider):

    '''
    Провайдер для модели
    '''
    def reader(self, grouped, offset, level_index, level_keys, begin, end, aggregates, sorting):
        return self.__read_model(grouped, offset, level_index, level_keys, begin, end, aggregates, sorting)

    def __read_model(self, grouped, offset, level_index, level_keys, begin, end, aggregates, sorting):
        '''
        вывод элементов дерева группировок в зависимости от уровня, ключевых элементов и интервала в уровне
        '''
        # построим ключ кэша
        # cache_key = '%s__%s__%s__%s__%s__%s' % (','.join(grouped), offset, level_index, ','.join(level_keys), begin, end)
        # if cache_key in out_cache.keys():
        #    print 'cached data...........'
        #    return out_cache[cache_key]

        # специальный режим, когда считается общий итог по всем записям - не важна сортировка и группировка
        if level_index == -1:
            aggr = []
            # будем считать агрегаты
            for agg in aggregates.keys():
                agg_type = aggregates[agg]
                if agg_type == 'sum':
                    aggr.append(Sum(agg))
                elif agg_type == 'count':
                    aggr.append(Count(agg))
                elif agg_type == 'min':
                    aggr.append(Min(agg))
                elif agg_type == 'max':
                    aggr.append(Max(agg))
                elif agg_type == 'avg':
                    aggr.append(Avg(agg))
            i = self.get_data().aggregate(*aggr)
            item = self.create_record()
            item.is_leaf = False
            item.index = None
            item.id = None
            item.indent = None
            item.lindex = None
            for agg in aggregates.keys():
                agg_type = aggregates[agg]
                if agg_type == 'sum':
                    setattr(item, agg, i[agg + '__sum'])
                elif agg_type == 'count':
                    setattr(item, agg, i[agg + '__count'])
                elif agg_type == 'min':
                    setattr(item, agg, i[agg + '__min'])
                elif agg_type == 'max':
                    setattr(item, agg, i[agg + '__max'])
                elif agg_type == 'avg':
                    setattr(item, agg, i[agg + '__avg'])
            item.calc()
            return item

        # print 'read_model(): grouped=%s, offset=%s, level_index=%s,
        # level_keys=%s, begin=%s, end=%s' % (grouped, offset, level_index,
        # level_keys, begin, end)
        res = []
        if grouped:
            # для всех группировочных элементов будут использоваться ключи
            # если берется уровень больший, чем количество группировок, то выдаем просто записи
            if level_index >= len(grouped):
                field = None
            else:
                field = grouped[level_index]

            filter = None
            for lev in range(0, level_index):
                lev_field = grouped[lev]
                key = level_keys[lev]
                if filter:
                    filter = filter & Q(**{lev_field: key})
                else:
                    filter = Q(**{lev_field: key})
            aggr = []
            if field:
                # будем считать агрегаты
                for agg in aggregates.keys():
                    agg_type = aggregates[agg]
                    if agg_type == 'sum':
                        aggr.append(Sum(agg))
                    elif agg_type == 'count':
                        aggr.append(Count(agg))
                    elif agg_type == 'min':
                        aggr.append(Min(agg))
                    elif agg_type == 'max':
                        aggr.append(Max(agg))
                    elif agg_type == 'avg':
                        aggr.append(Avg(agg))
            if filter:
                if field:
                    if aggr:
                        query = self.get_data().filter(filter).values(
                            field).annotate(*aggr).annotate(count=Count("id"))
                    else:
                        query = self.get_data().filter(filter).values(
                            field).annotate(count=Count("id"))
                else:
                    query = self.get_data().filter(filter)
            else:
                if field:
                    if aggr:
                        query = self.get_data().values(field).annotate(
                            *aggr).annotate(count=Count("id"))
                    else:
                        query = self.get_data().values(
                            field).annotate(count=Count("id"))
                else:
                    query = self.get_data()

            # сортировка
            sort_fields = []
            # оставлено для совместимости
            if isinstance(sorting, dict) and sorting.keys():
                for sort_field, sort_dir in sorting.iteritems():
                    #необходимо исключить из сортировки поля, которые не входят в aggr, иначе по ним будет сделана ненужная группировка
                    if sort_field in aggregates.keys() or sort_field == field or not field:
                        if sort_field in aggregates.keys() and field:
                            # необходимо добавить суффикс к полю, которое в аггрегатах, иначе сортировать будет не по тому что надо :)
                            sort_field = sort_field + '__' + aggregates[sort_field]
                        if sort_dir == 'DESC':
                            sort_fields.append('-' + sort_field)
                        else:
                            sort_fields.append(sort_field)
            # изначальная ошибка была передавать сортировку в dict - надо было сразу в list
            elif isinstance(sorting, list) and sorting:
                for sort_field, sort_dir in sorting:
                    #необходимо исключить из сортировки поля, которые не входят в aggr, иначе по ним будет сделана ненужная группировка
                    if sort_field in aggregates.keys() or sort_field == field or not field:
                        if sort_field in aggregates.keys() and field:
                            # необходимо добавить суффикс к полю, которое в аггрегатах, иначе сортировать будет не по тому что надо :)
                            sort_field = sort_field + '__' + aggregates[sort_field]
                        if sort_dir == 'DESC':
                            sort_fields.append('-' + sort_field)
                        else:
                            sort_fields.append(sort_field)
            else:
                # нет заданной сортировки, отсортируем по этому полю
                if field:
                    sort_fields.append(field)
            if sort_fields:
                query = query.order_by(*sort_fields)

            # теперь выведем запрошенные элементы уровня
            index = 0
            for i in query.all()[begin:end + 1]:
                if field:
                    item = self.create_record()
                    item.is_leaf = False
                    item.index = offset + index + begin
                    item.id = i[field]
                    item.indent = level_index
                    item.lindex = index + begin
                    item.count = i['count']
                    # установим все атрибуты из группировки
                    for lev in range(0, level_index):
                        lev_field = grouped[lev]
                        key = level_keys[lev]
                        setattr(item, lev_field, key)
                    setattr(item, field, i[field])

                    for agg in aggregates.keys():
                        agg_type = aggregates[agg]
                        if agg_type == 'sum':
                            setattr(item, agg, i[agg + '__sum'])
                        elif agg_type == 'count':
                            setattr(item, agg, i[agg + '__count'])
                        elif agg_type == 'min':
                            setattr(item, agg, i[agg + '__min'])
                        elif agg_type == 'max':
                            setattr(item, agg, i[agg + '__max'])
                        elif agg_type == 'avg':
                            setattr(item, agg, i[agg + '__avg'])
                    item.calc()
                else:
                    item = self.create_record()
                    item.is_leaf = True
                    item.index = offset + index + begin
                    item.indent = level_index
                    item.lindex = index + begin
                    item.load(i)
                    item.calc()
                res.append(item)
                index = index + 1
        else:
            # вывести без группировки
            index = 0
            query = self.get_data()
            # оставлено для совместимости
            if isinstance(sorting, dict) and sorting.keys():
                order = []
                for sort_field, sort_dir in sorting.iteritems():
                    if sort_dir == 'DESC':
                        order.append('-' + sort_field)
                    else:
                        order.append(sort_field)
                query = query.order_by(*order)
            # изначальная ошибка была передавать сортировку в dict - надо было сразу в list
            elif isinstance(sorting, list) and sorting:
                order = []
                for sort_field, sort_dir in sorting:
                    if sort_dir == 'DESC':
                        order.append('-' + sort_field)
                    else:
                        order.append(sort_field)
                query = query.order_by(*order)
            for i in query.all()[begin:end + 1]:
                item = self.create_record()
                item.indent = 0
                item.is_leaf = True
                item.count = 0
                item.lindex = index + begin
                item.index = index + begin
                item.load(i)
                item.calc()
                res.append(item)
                index = index + 1
        # out_cache[cache_key] = (res,total_of_level)
        return res

    def counter(self, grouped, level_index, level_keys, expandedItems):
        return self.__count_model(grouped, level_index, level_keys, expandedItems)

    def __count_model(self, grouped, level_index, level_keys, expandedItems):
        '''
        подсчет количества строк в раскрытом уровне
        '''
        # построим ключ кэша
        # cache_key = '%s__%s__%s__%s' % (','.join(grouped), level_index, ','.join(level_keys), add_to_count_key(expandedItems))
        # if cache_key in count_cache.keys():
        #    print 'cached count...........'
        #    return count_cache[cache_key]

        # print 'count_model(): grouped=%s, level_index=%s, level_keys=%s,
        # expandedItems=%s' % (grouped, level_index, level_keys, expandedItems)
        total_of_level = 0
        if grouped:
            grouped_ranges = []
            # определим порядок группировки
            for i in grouped:
                grouped_ranges.append(i)

            query = self.get_data()
            filter = None
            for lev in range(0, level_index):
                lev_field = grouped_ranges[lev]
                key = level_keys[lev]
                if filter:
                    filter = filter & Q(**{lev_field: key})
                else:
                    filter = Q(**{lev_field: key})
            if filter:
                query = query.filter(filter)
            if level_index < len(grouped_ranges):
                field = grouped_ranges[level_index]
                query = query.values(field).distinct()
                # т.к. count(distinct не считает поля с null, то будем считать ВЛОБ!
                total_of_level = len(query)  # query.count()
            else:
                total_of_level = query.count()

        else:
            total_of_level = self.get_data().count()

        # добавим к количеству также сумму раскрытых элементов
        exp_count = 0
        for exp in expandedItems:
            if exp['count'] == -1:
                exp['count'] = self.counter(grouped, level_index + 1, level_keys + [
                                            exp['id']], exp['expandedItems'])
            exp_count = exp_count + exp['count']

        # count_cache[cache_key] = total_of_level+exp_count

        # print 'count_model() = %s, total=%s, exp_count=%s' % (total_of_level
        # + exp_count, total_of_level, exp_count)
        return total_of_level + exp_count

    def indexer(self, grouped, level_index, level_keys, expandedItems, aggregates, sorting):
        return self.__index_model(grouped, level_index, level_keys, expandedItems, aggregates, sorting)

    def __index_model(self, grouped, level_index, level_keys, expandedItems, aggregates, sorting):
        '''
        построение индексов элементов в раскрытом уровне, только для группировок и для тех, которые раскрыты
        '''
        res = []
        if grouped and len(expandedItems) > 0:
            # для всех группировочных элементов будут использоваться ключи
            # если берется уровень больший, чем количество группировок, то выдаем просто записи
            field = grouped[level_index]

            filter = None
            for lev in range(0, level_index):
                lev_field = grouped[lev]
                key = level_keys[lev]
                if filter:
                    filter = filter & Q(**{lev_field: key})
                else:
                    filter = Q(**{lev_field: key})
            aggr = []
            # будем считать агрегаты - по ним тоже сортируют иногда
            for agg in aggregates.keys():
                agg_type = aggregates[agg]
                if agg_type == 'sum':
                    aggr.append(Sum(agg))
                elif agg_type == 'count':
                    aggr.append(Count(agg))
                elif agg_type == 'min':
                    aggr.append(Min(agg))
                elif agg_type == 'max':
                    aggr.append(Max(agg))
                elif agg_type == 'avg':
                    aggr.append(Avg(agg))

            query = self.get_data().values(field)
            if filter:
                query = query.filter(filter)
            if aggr:
                query = query.annotate(*aggr).annotate(count=Count("id"))
            else:
                query = query.distinct()

            # сортировка
            sort_fields = []
            # оставлено для совместимости
            if sorting:
                if isinstance(sorting, dict):
                    pairs = sorting.iteritems()
                elif isinstance(sorting, list):
                    pairs = sorting
                else:
                    raise TypeError("Sorting must be dict or list!")
                for sort_field, sort_dir in pairs:
                    # необходимо исключить из сортировки поля,
                    # которые не входят в aggr,
                    # иначе по ним будет сделана ненужная группировка
                    if sort_field in aggregates.keys() or sort_field == field:
                        if sort_field in aggregates.keys():
                            # необходимо добавить суффикс к полю,
                            # которое в аггрегатах,
                            # иначе сортировать будет не по тому что надо :)
                            sort_field = sort_field + '__' + (
                                aggregates[sort_field]
                            )
                        if sort_dir == 'DESC':
                            sort_fields.append('-' + sort_field)
                        else:
                            sort_fields.append(sort_field)
            else:
                # нет заданной сортировки, отсортируем по этому полю
                sort_fields.append(field)
            if sort_fields:
                query = query.order_by(*sort_fields)
            # теперь выведем запрошенные элементы уровня
            for i in query:
                res.append(i[field])
        return res


class GroupingRecordDataProvider(GroupingRecordProvider):
    '''
    Провайдер для массива
    '''
    def update_aggregates(self, aggregates, aggr_rec, rec):
        for agg in aggregates.keys():
            agg_type = aggregates[agg]
            agg_value = self.getattr(rec, agg)
            if agg_type == 'sum':
                aggr_rec[agg] = (
                    (agg_value + aggr_rec[agg])
                    if agg in aggr_rec else
                    agg_value
                )
            elif agg_type == 'count':
                aggr_rec[agg] = 1 + aggr_rec.get(agg, 0)
            elif agg_type == 'min':
                aggr_rec[agg] = min(
                    agg_value,
                    aggr_rec.get(agg, agg_value)
                )
            elif agg_type == 'max':
                aggr_rec[agg] = max(
                    agg_value,
                    aggr_rec.get(agg, agg_value)
                )
            elif agg_type == 'avg':
                aggr_rec[agg] = (
                    (agg_value + aggr_rec[agg])
                    if agg in aggr_rec else
                    agg_value
                )
            elif callable(agg_type):
                aggr_rec[agg] = agg_type(aggr_rec.get(agg), agg_value)

    def reader(
        self, grouped, offset, level_index, level_keys,
        begin, end, aggregates, sorting
    ):
        '''
        вывод элементов дерева группировок в зависимости от уровня,
        ключевых элементов и интервала в уровне
        '''
        # специальный режим, когда считается общий итог по всем записям
        # - не важна сортировка и группировка
        if level_index == -1:
            aggr_rec = {}
            count = 0
            for rec in self.get_data():
                # будем считать агрегаты
                self.update_aggregates(aggregates, aggr_rec, rec)
                count += 1
            item = self.create_record()
            self.setattr(item, 'id', None)
            self.setattr(item, 'indent', None)
            self.setattr(item, 'lindex', None)
            self.setattr(item, 'count', count)
            if aggr_rec:
                for agg, aggv in aggregates.iteritems():
                    # для средних - посчитаем среднее
                    if aggv == 'avg':
                        self.setattr(
                            item, agg,
                            aggr_rec[agg] / self.getattr(item, 'count')
                        )
                    elif callable(aggv):
                        self.setattr(
                            item, agg,
                            aggv(aggr_rec[agg], None)
                        )
                    else:
                        self.setattr(item, agg, aggr_rec[agg])
            self.calc(item)
            return item

        # сортировать будем после группировки и фильтрации
        raw_data = self.get_data()
        res = []
        pre_res = []

        if grouped:
            # для всех группировочных элементов будут использоваться ключи
            level = {}
            aggregate_values = {}
            prepared = []

            # если берется уровень больший, чем количество группировок,
            # то выдаем просто записи
            if level_index >= len(grouped):
                field = None
            else:
                field = grouped[level_index]

            for rec in raw_data:
                found = True
                for lev in range(0, level_index):
                    lev_field = grouped[lev]
                    key = level_keys[lev]
                    key_value = self.getattr(rec, lev_field)
                    # подходит ли запись под группировку
                    if key != key_value:
                        found = False
                        break
                # если успешно проверили все поля, то значит это наша запись
                if found:
                    if field:
                        group_value = self.getattr(rec, field)
                        if not group_value in level.keys():
                            level[group_value] = 1
                            aggr_rec = {}
                            aggregate_values[group_value] = aggr_rec
                            prepared.append((group_value, rec))
                        else:
                            level[group_value] += 1
                            aggr_rec = aggregate_values[group_value]
                        # будем считать агрегаты
                        self.update_aggregates(aggregates, aggr_rec, rec)
                    else:
                        prepared.append((rec, rec))
            # теперь выведем запрошенные элементы уровня
            # придется обработать все записи уровня,
            # т.к. требуется еще отсортировать их,
            # и лишь потом ограничить количество
            for i, rec in prepared:
                if field:
                    item = self.create_record(grouped=grouped)
                    self.setattr(item, field, i)
                    self.setattr(item, 'id', i)
                    self.setattr(item, 'indent', level_index)
                    self.setattr(item, 'count', level[i])
                    for agg, aggv in aggregates.iteritems():
                        # для средних - посчитаем среднее
                        if agg == 'avg':
                            self.setattr(
                                item, agg,
                                aggregate_values[i][agg] / (
                                    self.getattr(item, 'count'))
                            )
                        elif callable(aggv):
                            self.setattr(
                                item, agg,
                                aggv(aggregate_values[i][agg], None)
                            )
                        else:
                            self.setattr(item, agg, aggregate_values[i][agg])

                    # Добавляем в прокси расшифровку для группируемого поля
                    detail_attr = self.detail_attrs_map.get(field)
                    if detail_attr:
                        self.setattr(item, detail_attr, self.getattr(
                            rec, detail_attr))

                    # проставим значения ключей уровня
                    for lev in range(0, level_index):
                        lev_field = grouped[lev]
                        key = level_keys[lev]
                        self.setattr(item, lev_field, key)

                        # Добавляем в прокси атрибуты для расшифровки
                        detail_attr = self.detail_attrs_map.get(lev_field)
                        if detail_attr:
                            self.setattr(item, detail_attr, self.getattr(
                                rec, detail_attr))

                    self.calc(item)
                else:
                    item = self.create_record(grouped=grouped)
                    self.setattr(item, 'is_leaf', True)
                    self.setattr(item, 'indent', level_index)
                    self.load(item, i)
                    self.calc(item)
                pre_res.append(item)

        else:
            # вывести все записи без группировки
            index = 0
            for i in raw_data:
                item = self.create_record()
                self.setattr(item, 'indent', 0)
                self.setattr(item, 'is_leaf', True)
                self.load(item, i)
                self.calc(item)
                pre_res.append(item)

        # а вот теперь сортируем и граничиваем
        # оставлено для совместимости
        if isinstance(sorting, dict) and sorting.keys():
            order = []
            for sort_field, sort_dir in sorting.iteritems():
                if sort_dir == 'DESC':
                    order.append('-' + sort_field)
                else:
                    order.append(sort_field)
            sorted_data = self.multikeysort(pre_res, order)
        # изначальная ошибка была передавать сортировку в dict
        # - надо было сразу в list
        elif isinstance(sorting, list) and sorting:
            order = []
            for sort_field, sort_dir in sorting:
                if sort_dir == 'DESC':
                    order.append('-' + sort_field)
                else:
                    order.append(sort_field)
            sorted_data = self.multikeysort(pre_res, order)
        else:
            sorted_data = pre_res
        index = 0
        for item in sorted_data[begin:end + 1]:
            self.setattr(item, 'index', offset + index + begin)
            self.setattr(item, 'lindex', index + begin)
            res.append(item)
            index += 1

        return res

    __read_data = reader

    def counter(self, grouped, level_index, level_keys, expandedItems):
        '''
        подсчет количества строк в раскрытом уровне
        '''
        total_of_level = 0
        if grouped:
            if level_index == 0:
                # вывести элементы 1-го уровня группировки
                # (не нужно использовать ключи)
                level = []
                # переберем элементы и сформируем уровень
                field = grouped[level_index]
                for rec in self.get_data():
                    group_value = self.getattr(rec, field)
                    if not group_value in level:
                        level.append(group_value)
                total_of_level = len(level)
            else:
                # для всех остальных элементов будут использоваться ключи
                level = []
                # если берется уровень больший, чем количество группировок,
                # то выдаем просто записи
                if level_index >= len(grouped):
                    field = None
                else:
                    field = grouped[level_index]

                for rec in self.get_data():
                    for lev in range(0, level_index):
                        lev_field = grouped[lev]
                        key = level_keys[lev]
                        key_value = self.getattr(rec, lev_field)
                        # подходит ли запись под группировку
                        if key != key_value:
                            break
                        # если успешно проверили все поля,
                        # то значит это наша запись
                        elif lev == level_index - 1:
                            if field:
                                group_value = self.getattr(rec, field)
                                if not group_value in level:
                                    level.append(group_value)
                            else:
                                level.append(rec)
                total_of_level = len(level)
        else:
            total_of_level = len(self.get_data())

        # добавим к количеству также сумму раскрытых элементов
        exp_count = 0
        for exp in expandedItems:
            if exp['count'] == -1:
                exp['count'] = self.counter(
                    grouped,
                    level_index + 1,
                    level_keys + [exp['id']],
                    exp['expandedItems']
                )
            exp_count = exp_count + exp['count']

        # count_cache[cache_key] = total_of_level+exp_count
        # print 'count_exp_data() = %s, total=%s, exp_count=%s' %
        # (total_of_level+exp_count, total_of_level, exp_count)
        return total_of_level + exp_count

    __count_data = counter

    def indexer(
        self, grouped, level_index, level_keys,
        expandedItems, aggregates, sorting
    ):
        '''
        построение индексов элементов в раскрытом уровне,
        только для группировок и для тех, которые раскрыты
        '''
        res = []
        pre_res = []
        if grouped and len(expandedItems) > 0:
            raw_data = self.get_data()
            # для всех группировочных элементов будут использоваться ключи
            level = {}
            aggregate_values = {}
            prepared = []
            field = grouped[level_index]

            for rec in raw_data:
                finded = True
                for lev in range(0, level_index):
                    lev_field = grouped[lev]
                    key = level_keys[lev]
                    key_value = self.getattr(rec, lev_field)
                    # подходит ли запись под группировку
                    if key != key_value:
                        finded = False
                        break
                # если успешно проверили все поля, то значит это наша запись
                if finded:
                    group_value = self.getattr(rec, field)
                    if not group_value in level.keys():
                        level[group_value] = 1
                        aggr_rec = {}
                        aggregate_values[group_value] = aggr_rec
                        prepared.append((group_value, rec))
                    else:
                        level[group_value] = level[group_value] + 1
                        aggr_rec = aggregate_values[group_value]
                    # будем считать агрегаты
                    self.update_aggregates(aggregates, aggr_rec, rec)

            # придется обработать все записи уровня,
            # т.к. требуется еще отсортировать их
            # и лишь потом ограничить количество
            for i, rec in prepared:
                item = self.create_record()
                self.setattr(item, field, i)
                self.setattr(item, 'id', i)
                self.setattr(item, 'indent', level_index)
                self.setattr(item, 'count', level[i])
                for agg in aggregates.keys():
                    # для средних - посчитаем среднее
                    if aggregates[agg] == 'avg':
                        self.setattr(
                            item, agg, (
                                aggregate_values[i][agg] /
                                self.getattr(item, 'count')))
                    else:
                        self.setattr(item, agg, aggregate_values[i][agg])

                # Добавляем в прокси расшифровку для группируемого поля
                detail_attr = self.detail_attrs_map.get(field)
                if detail_attr:
                    self.setattr(
                        item, detail_attr, self.getattr(rec, detail_attr))

                # проставим значения ключей уровня
                for lev in range(0, level_index):
                    lev_field = grouped[lev]
                    key = level_keys[lev]
                    self.setattr(item, lev_field, key)

                    # Добавляем в прокси атрибуты для расшифровки
                    detail_attr = self.detail_attrs_map.get(lev_field)
                    if detail_attr:
                        self.setattr(
                            item, detail_attr, self.getattr(rec, detail_attr))

                self.calc(item)
                pre_res.append(item)

            # а вот теперь сортируем и граничиваем
            # оставлено для совместимости
            if isinstance(sorting, dict) and sorting.keys():
                order = []
                for sort_field, sort_dir in sorting.iteritems():
                    if sort_dir == 'DESC':
                        order.append('-' + sort_field)
                    else:
                        order.append(sort_field)
                sorted_data = self.multikeysort(pre_res, order)
            # изначальная ошибка была передавать сортировку в dict
            # - надо было сразу в list
            elif isinstance(sorting, list) and sorting:
                order = []
                for sort_field, sort_dir in sorting:
                    if sort_dir == 'DESC':
                        order.append('-' + sort_field)
                    else:
                        order.append(sort_field)
                sorted_data = self.multikeysort(pre_res, order)
            else:
                sorted_data = pre_res

            for item in sorted_data:
                res.append(self.getattr(item, 'id'))
        return res

    __index_data = indexer

    def multikeysort(self, items, columns):
        comparers = [
            (
                (col[1:].strip(), -1)
                if col.startswith('-') else
                (col.strip(), 1)
            ) for col in columns
        ]
        def comparer(left, right):
            for attr, mult in comparers:
                result = cmp(
                    self.getattr(left, attr),
                    self.getattr(right, attr)
                )
                if result:
                    return mult * result
            else:
                return 0
        return sorted(items, cmp=comparer)



#coding: utf-8
'''
Created on 14.04.2010

@author: airat
'''

import decimal
import time
import datetime
import re
import codecs
import uuid
import string
import xml.dom.minidom

from django.utils.encoding import smart_unicode
from django.db import models, transaction

try:
    from dbfpy.dbf import Dbf
except ImportError:
    # Заглушка на случай неустановленного dbfpy
    class Dbf(object):
        def __new__(cls, *args, **kwargs):
            raise RuntimeError(u'dbfpy is not installed!')


class DictNotEmptyException(Exception):
    '''
    Эксепшн, который выбрасывается в случае, если заполняемый справочник не пуст.
    '''
    def __init__(self, model, reason):
        # указывает на модель записи справочника, для которой определилось, что есть записи в базе данных
        self.model = model
        # текстовое описание причины исключения
        self.reason = reason

class DictLoadException(Exception):
    '''
    Эксепшн, который вызывается, если во время загрузки справочника из файла
    возникла какая-либо ошибка.
    '''
    def __init__(self, model, reason, orig_exc_info=None):
        # указывает на модель записи справочника, при загрузки которой возникла ошибка
        # в некоторых случаях(при считывании данных из файла) может быть == None
        self.model = model
        # текстовое описание причины исключения
        self.reason = reason
        #TODO нужно что-то еще сделать с исходным эксепшеном. хотя бы прокинуть дальше value и traceback

def read_simple_CSV_dict_file(filename, comma=','):
    '''
        Чтение данных из файла (формата CSV), где первая строка содержит перечисление
        имен атрибутов, а последующие строки содержат значения для заданных атрибутов для
        каждой записи в порядке, указанном в первой строке.
        Струткура возвращаемого значения аналогична той, что с методом @see read_simple_dict_file(filename)
    '''
    assert isinstance(filename, basestring), u"filename must be 'str'"

    attrs = []
    values = []

    try:
        with open(filename, "rb") as f:
            first_string = f.readline()
            first_string = strip_bom(first_string)


            attrs = [item.strip().lower() for item in smart_unicode(first_string).split(comma)]
            for num, line in enumerate(f, 1):
                row_values = [item.strip() for item in smart_unicode(line).split(comma)]
                if len(row_values) == len(attrs):
                    values.append(row_values)
                else:
                    raise DictLoadException(None, u'Количество полей не совпадает: строка %s' % num)
    except IOError:
        raise

    return (attrs, values,)

def read_simple_DBF_dict_file(filename, encoding='utf-8',
                              code_field="CODE", name_field="NAME"):
    ''' Чтение данных из файла формата "DBase" (.dbf) '''

    assert isinstance(filename, basestring), u"filename must be 'str'"

    attrs = []
    values = []

    try:
        records = Dbf(filename, readOnly=True, new=False)
        if records:
            for k,v in records[0].asDict().items(): # каждая запись - это ассоц. массив, ключи - имена атрибутов
                attrs.append(k)

            for record in records:
                vals_here = []
                for name in attrs:
                    vals_here.append(smart_unicode(record[name], encoding=encoding))

                values.append(vals_here)

        else:
            raise DictLoadException(None, u'Исходные данные отсутствуют. Пустой DBF-файл')
    except IOError:
        raise
    except TypeError:
        raise

    attrs_lowercase = []
    for attr in attrs:
        if attr not in [code_field, name_field, ]:
            attrs_lowercase.append(attr.strip().lower())
        else:
            if attr == code_field:
                attrs_lowercase.append("code")
            else:
                attrs_lowercase.append("name")

    attrs = attrs_lowercase

    return (attrs, values,)

def read_simple_xml_file(filename, fields={}):
    '''
    Чтение данных из файла формата xml.
    fields: ключ - атрибут в узле, соответствуюший полю в таблице
    значение - поле в таблице
    '''
    assert isinstance(filename, basestring), u"filename must be basestring type"

    values = []
    attrs = []
    doc = xml.dom.minidom.parse(filename)
    root_node = doc.childNodes
    if root_node[0]:
        for node in root_node[0].childNodes:
            values_list = []
            for attr in node.attributes._attrs:
                if attr in fields:
                    if not fields[attr] in attrs:
                        attrs.append(fields[attr])
                    values_list.append(smart_unicode(node.attributes._attrs[attr].nodeValue, encoding=doc.encoding))
            values.append(values_list)
    else:
        raise DictLoadException(None, u'Исходные данные отсутствуют. Пустой XML-файл')

    return (attrs, values,)

def read_simple_dict_file(filename):
    '''
    Выполняет чтение данных из формата простого линейного справочника
    и возвращает записи в виде кортежа со списком значений справочника.
    Пример возвращаемого значения,
    ( ['code', 'name'],
      [ ['Код1','Наименование 1'],
        ['Код2','Наименование 2'],
      ]
    )
    '''

    assert isinstance(filename, basestring), u"filename must be 'str'"

    attrs = []
    values = []
    try:
        with open(filename, "rb") as f:
            first_string = f.readline()
            first_string = strip_bom(first_string)
            if first_string.startswith('#'):
                first_string = first_string[1:].strip()
            else:
                raise DictLoadException(None, u'Первая строка скрипта загрузки линейного справочника должна содержать название аттрибутов и начинаться с символа #')

            attrs = [item.strip() for item in smart_unicode(first_string).split('\t')]
            for num, line in enumerate(f, 1):
                row_values = [item.strip() for item in line.split('\t')]
                if len(row_values) == len(attrs):
                    values.append(row_values)
                else:
                    raise DictLoadException(None, u'Количество полей не совпадает: строка %s' % num)
    except IOError:
        raise

    return (attrs, values,)

def read_tree_dict_file(filename):
    '''
    Выполняет чтение данных из формата иерархического справочнка.
    Пример возвращаемого значение:
    ( ['name'],
      ['code', 'name'],
      [.... описание пока не завершено ..... ]
    )
    '''

    assert isinstance(filename, basestring), u"filename must be 'str'"

    # паттерн для строки модели групп
    VALID_BRANCH = '\[\+\](\w| )+'
    tree_attrs = []
    dict_attrs = []

    class StringStruct():
        '''
        Парсит строку: определяет уровень вложенности, значения аттрибутов,
        является ли запись веткой или листом, генерит уникальный id'шник
        '''

        def __init__(self, row_string):
            if row_string == '':
                raise DictLoadException(None ,u'Пустая строка')

            s = row_string.split('\t')
            # Делаем для каждой записи уникальный id'шник
            self.uid = str(uuid.uuid4())[0:16]
            self.parent_uid = None
            self.level = 0
            self.is_leaf = True
            # значения аттрибутов
            self.attrs = None
            for item in s:
                if item == '':
                    self.level += 1
                else:
                    # знак '-' - это обозначение пустого значения, чтобы отличать начало строки от уровней вложенности ветки
                    # (может использоваться только для первого элемента в строке)
                    if item == '-':
                        s[self.level] = ''
                    break
            self.attrs = [ item.strip() for item in s[self.level:]]
            if self.attrs:
                self.is_leaf = not re.match(VALID_BRANCH, self.attrs[0])
                # убираем плюсик в скобках у веток
                if not self.is_leaf:
                    self.attrs[0] = self.attrs[0][3:].strip()
                    if self.attrs[0] == '-':
                        self.attrs[0] = ' '

            else:
                # листьям разрешено не иметь полей
                self.is_leaf = True

    try:
        with open(filename, 'rb') as f:
            lines = f.readlines()
            if len(lines) < 3:
                raise DictLoadException(None, u'В файле должно быть, как минимум, три строки')

            first_string = strip_bom(lines[0])
            if first_string.startswith('#'):
                first_string = first_string[1:].strip()
            else:
                raise DictLoadException(None, u'Первая строка скрипта загрузки иерархического справочника должна содержать название модели групп и начинаться с символа #')

            second_string = lines[1]
            if second_string.startswith('#'):
                second_string = second_string[1:].strip()
            else:
                raise DictLoadException(None, u'Вторая строка скрипта загрузки иерархического справочника должна содержать название аттрибутов модели линейного справочника и начинаться с символа #')

            tree_attrs = [item.strip() for item in smart_unicode(first_string).split('\t')]
            dict_attrs = [item.strip() for item in smart_unicode(second_string).split('\t')]
            dict_rows = {}
            tree_rows = {}
            parents = []
            cur_parent = -1
            level = 0
            for num, line in enumerate(lines[2:], 3):
                #пустые строки в файле пропускаются
                if line == '':
                    continue
                struct = StringStruct(line)
                if (struct.level == level) or (struct.level-1 == level):
                    # все нормально, тут ничего не делаем
                    pass
                elif struct.level < level:
                    d = level - struct.level
                    for i in range(d):
                        cur_parent = parents.pop()
                    level -= d
                else:
                    raise DictLoadException(None, u'Уровень вложенности задан неверно: строка %s' % num)

                struct.parent_uid = cur_parent
                if struct.is_leaf:
                    dict_rows[struct.uid] = struct
                else:
                    parents.append(cur_parent)
                    cur_parent = struct.uid
                    level += 1
                    tree_rows[struct.uid] = struct

    except IOError:
        raise

    return (tree_attrs, dict_attrs, tree_rows, dict_rows)

@transaction.commit_on_success
def fill_simple_dict(model, data):
    '''
    Выполняет заполнение модели model (тип параметра - classobj) данными data,
    в формате, возвращаемом функцией read_simple_dict_file
    '''

    assert issubclass(model, models.Model), 'model must be subclass of django.db.models.Model'

    if model.objects.all().count() > 0:
        raise DictNotEmptyException(model.__name__, u'Таблица справочника %s не должна содержать записи' % model.__name__)

    fields = dict( (field.name, field,) for field in model._meta.fields)
    attrs = data[0]
    values = data[1]
    for value in values:
        obj = model()
        for attr, val in zip(attrs, value):
            if val != u'None':
                try:
                    if fields.has_key(attr):
                        converted = _convert_value(fields[attr], val)
                        setattr(obj, attr, converted)
                except Exception as e:
                    raise DictLoadException(model.__name__, u'Не удалось преобразовать значение: %s' % val)
        try:
            obj.save()
        except Exception as exc:
            raise DictLoadException(model.__name__, u'Не удалось сохранить запись справочника: %s' % string.join(value))

@transaction.commit_on_success
def fill_tree_dict(group_model, list_model, group_link, list_link, data):

    assert issubclass(list_model, models.Model), 'model must be subclass of django.db.models.Model'
    assert issubclass(group_model, models.Model), 'model must be subclass of django.db.models.Model'

    if group_model.objects.all().count() > 0:
        raise DictNotEmptyException(group_model.__name__, u'Таблица справочника %s не должна содержать записи' % group_model.__name__)

    if list_model.objects.all().count() > 0:
        raise DictNotEmptyException(list_model.__name__, u'Таблица справочника %s не должна содержать записи' % list_model.__name__)

    tree_attrs = data[0]
    dict_attrs = data[1]
    tree_rows = data[2]
    dict_rows = data[3]
    tree_values = {}

    tree_fields = dict( (field.name, field,) for field in group_model._meta.fields)
    for k,v in tree_rows.items():
        obj = group_model()
        for i in range(len(tree_attrs)):
            try:
                val = _convert_value(tree_fields[str(tree_attrs[i])], v.attrs[i])
            except:
                raise DictLoadException(group_model.__name__, u'Не удалось преобразовать значение: %s' % v.attrs[i])
            setattr(obj, str(tree_attrs[i]), val)
        try:
            obj.save()
        except:
            raise DictLoadException(group_model.__name__, u'Не удалось сохранить запись справочника: %s' % string.join(v.attrs))
        tree_values[k] = (obj, v.parent_uid,)

    for k,v in tree_values.items():
        if v[1] != -1:
            setattr(v[0], group_link, tree_values[v[1]][0])
            try:
                v[0].save()
            except:
                raise DictLoadException(group_model.__name__, u'Не удалось сохранить запись справочника: %s' % string.join(v[0]))

    dict_fields = dict( (field.name, field,) for field in list_model._meta.fields)
    for k,v in dict_rows.items():
        obj = list_model()
        for i in range(len(dict_attrs)):
            if v.attrs[i] != u'None':
                try:
                    val = _convert_value(dict_fields[str(dict_attrs[i])], v.attrs[i])
                except:
                    raise DictLoadException(list_model.__name__, u'Не удалось преобразовать значение: %s' % v.attrs[i])
                setattr(obj, str(dict_attrs[i]), val)

        parent = None if v.parent_uid == -1 else tree_values[v.parent_uid][0]
        setattr(obj, list_link, parent)
        try:
            obj.save()
        except:
            raise DictLoadException(list_model.__name__, u'Не удалось сохранить запись справочника: %s' % string.join(v.attrs, ','))

def strip_bom(s):
    '''
    Убирает от начала строки все символы BOM
    '''
    boms = (codecs.BOM, codecs.BOM32_BE, codecs.BOM32_LE, codecs.BOM64_BE,
            codecs.BOM64_LE, codecs.BOM_BE, codecs.BOM_LE, codecs.BOM_UTF16,
            codecs.BOM_UTF16_BE, codecs.BOM_UTF16_LE, codecs.BOM_UTF32,
            codecs.BOM_UTF32_BE, codecs.BOM_UTF32_LE, codecs.BOM_UTF8,)
    return s.lstrip(''.join(boms))

def _convert_value(field, value):
    if not value:
        return None
    converted_value = value
    if isinstance(field, models.AutoField):
        pass
    elif isinstance(field, models.BooleanField):
        value = value.lower()
        if value == 'true' or value == '1':
            converted_value = True
        elif value == 'false' or value == '0':
            converted_value = False
        else:
            raise Exception()
    elif isinstance(field, models.CharField):
        pass
    elif isinstance(field, models.CommaSeparatedIntegerField):
        pass
    elif isinstance(field, models.DateTimeField):
        if '-' in value:
            ts = time.strptime(value, '%Y-%m-%d %H.%M.%S')
        else:
            ts = time.strptime(value, '%d.%m.%Y %H.%M.%S')
        converted_value = datetime.datetime(*(ts[0:6]))
    elif isinstance(field, models.DateField):
        if '-' in value:
            ts = time.strptime(value, '%Y-%m-%d')
        else:
            ts = time.strptime(value, '%d.%m.%Y')
        converted_value = datetime.date(*(ts[0:3]))
    elif isinstance(field, models.DecimalField):
        converted_value = decimal.Decimal(value)
    elif isinstance(field, models.EmailField):
        pass
    elif isinstance(field, models.FileField):
        pass
    elif isinstance(field, models.FilePathField):
        pass
    elif isinstance(field, models.FloatField):
        converted_value = float(value.replace(',', '.'))
    elif isinstance(field, models.ImageField):
        pass
    elif isinstance(field, models.IntegerField):
        converted_value = int(value)
    elif isinstance(field, models.IPAddressField):
        pass
    elif isinstance(field, models.NullBooleanField):
        value = value.lower()
        if value == 'true':
            converted_value = True
        elif value == 'false':
            converted_value = False
        elif value in ['null', 'none', '']:
            converted_value = None
    elif isinstance(field, models.PositiveIntegerField):
        converted_value = int(value)
    elif isinstance(field, models.PositiveSmallIntegerField):
        converted_value = int(value)
    elif isinstance(field, models.SlugField):
        pass
    elif isinstance(field, models.SmallIntegerField):
        converted_value = int(value)
    elif isinstance(field, models.TextField):
        pass
    elif isinstance(field, models.TimeField):
        ts = time.strptime(value, '%H.%M.%S')
        converted_value = datetime.time(*(ts[3:3]))
    elif isinstance(field, models.URLField):
        pass
    elif isinstance(field, models.XMLField):
        pass

    return converted_value



#coding: utf-8
'''
Хелперы для валидации значений
Created on 20.01.2010

@author: akvarats
'''
from django.db import models

class Validator:
    '''
    Класс, обеспечивающий выполнение валидации объектов
    на корректность их состояния.
    '''
    def __init__(self):
        self.rules = []
    
#------------------------------------------------------------------------------ 
# методы добавления правил валидации
#------------------------------------------------------------------------------ 
    def addrule_notempty(self, value, fail_msg, on_success=None, on_fail=None):
        '''
        Добавляет правило валидации значения на непустоту
        ''' 
        rule = NotEmptyValidationRule()
        rule.value = value
        rule.fail_msg = fail_msg
        rule.callback_on_success = on_success
        rule.callback_on_fail = on_fail
        self.rules.append(rule)
        
    def addrule_failed(self, fail_msg, on_success=None, on_fail=None):
        rule = FailedValidationRule()
        rule.fail_msg = fail_msg
        rule.callback_on_success = on_success
        rule.callback_on_fail = on_fail
        self.rules.append(rule)
        
    def addrule_equal(self, value, fail_msg, on_success=None, on_fail=None):
        rule = NotEqualValidationRule()
        rule.value = value
        rule.fail_msg = fail_msg
        rule.callback_on_success = on_success
        rule.callback_on_fail = on_fail
        self.rules.append(rule)
        
    
    def addrule_unique(self, value, model, field_name,fail_msg, on_success=None, 
                       on_fail=None):
        '''
        Добавляет правило валидации на уникальность заданного поля
        '''
        rule = ModelUniqueField(value, model, field_name)
        rule.value = value
        rule.fail_msg = fail_msg
        rule.callback_on_success = on_success
        rule.callback_on_fail = on_fail
        self.rules.append(rule)
    
        
    #------------------------------------------------------------------------------
    # методы добавления правил валидации 
    #------------------------------------------------------------------------------ 
    def check(self, raise_exception=True):
        result = []
        for rule in self.rules:
            if not rule.check():
                result.append(rule.fail_msg)
                if rule.callback_on_fail:
                    rule.callback_on_fail()
            else:
                if rule.callback_on_success:
                    rule.callback_on_success()
        if result and raise_exception:
            e = ValidationFailed()
            for message in result:
                e.add_error_message(message)
            raise e
            
        return result
        
class BaseValidationRule(object):
    '''
    Базовое правило валидации значений.
    
    Ничего по сути не делает
    '''
    def __init__(self):
        self.value = ''
        self.fail_msg = '' 
        self.callback_on_success = None
        self.callback_on_fail = None
    def check(self):
        return True
        
class NotEmptyValidationRule(BaseValidationRule):
    '''
    Правило валидации значения на непустоту
    '''
    def check(self):
        return True if self.value else False
    
class NotEqualValidationRule(BaseValidationRule):
    '''
    Правило проверяет равны ли значения из списка value между собой.
    Например может пригодиться для сравнения пароля и подтверждения 
    '''
    def check(self):
        assert isinstance(self.value, (list, tuple))
        assert len(self.value) >= 2
        last_value = self.value[0]
        for value in self.value[1:]:
            if last_value != value:
                return False
            last_value = value
        return True

class ModelUniqueField(BaseValidationRule):
    '''
    Проверяет что value нет ни у одной записи в модели.
    Например у нас есть пользователь с мылом и мы хотим убедиться что такого же мыла 
    нет ни у кого больше. Причем проверяемая запись не исключается. 
    '''
    def __init__(self, value, model, field_name):
        super(ModelUniqueField, self).__init__()
        self.value = value
        self.model = model
        self.field_name = field_name
    
    def check(self):
        assert issubclass(self.model, models.Model)
        assert isinstance(self.field_name, str)
        query = self.model.objects.filter(**{self.field_name: self.value})
        if isinstance(self.value, models.Model):
            if self.value.id != None:
                query = query.exclude(id = self.value.id)
        return not query.exists()

class FailedValidationRule(BaseValidationRule):
    '''
    Правило валидации, которое заведомо является неправильным
    '''
    def check(self):
        return False # надо тупо записать текущее сообщение в список ошибок

#===============================================================================
# Исключительная ситуация 
#===============================================================================
class ValidationFailed(Exception):
    '''
    Исключение для ситуаций валидации пользовательских данных
    '''
    def __init__(self, *args):
        Exception.__init__(self, *args)
        self.error_messages = [unicode(arg) for arg in args] # список ошибок валидации
        
    def add_error_message(self, message):
        self.error_messages.append(message)

    def has_errors(self):
        return self.error_messages

    def __unicode__(self):
        return self.messages_as_string()

    def __str__(self):
        return unicode(self).encode("utf-8")

    def messages_as_string(self, line_preffix = '  - ', line_delimiter='\n'):
        '''
        Преобразует self.error_messages в одну строку для отдачи в html
        представление
        '''
        result = ''
        for message in self.error_messages:
            result += line_preffix + message + line_delimiter
        if result:
            result = result[0:len(result)-len(line_delimiter)]
        return result

    def messages_to_html(self):
        return self.messages_as_string(line_preffix = '&nbsp;&nbsp;-&nbsp;', line_delimiter='<br>')



#coding:utf-8
'''
Created on 29.03.2010

@author: prefer
'''
import copy

class TypedList(list):
    '''
        Вспомогательный класс, перекрывает три метода - добавление (append, insert, extend), изменение и расширение списка
        Атрибут должен быть типом, переданном в качестве параметра type, или соответствовать значению,
        переданному в *aprgs - иначе выводится assert
        Если значение входит в список исключений exceptions, то проверка не производится
    '''
    def __init__(self, type, exceptions=[], on_before_addition=None,
                 on_after_addition=None, on_before_deletion=None,
                 on_after_deletion=None):
        ''' 
        @param type: Тип значений, которые должны находится в списке
        '''
        super(TypedList, self).__init__()
        self._type = type
        self._exceptions = exceptions
        self.on_before_addition = on_before_addition
        self.on_after_addition = on_after_addition
        self.on_before_deletion = on_before_deletion
        self.on_after_deletion = on_after_deletion
        
    def __copy__(self):
        """ Правильное клонирование TypedList, которое copy не может сделать автоматически """
        result = TypedList(
            type = self._type, 
            exceptions = self._exceptions,
            on_before_addition = self.on_before_addition,
            on_after_addition = self.on_after_addition,
            on_before_deletion = self.on_before_deletion,
            on_after_deletion = self.on_after_deletion
        )
        result.extend(self[:])
        return result

    def __deepcopy__(self, memo):
        """ Глубокое клонирование TypedList, которое copy не может сделать автоматически """
        result = TypedList(
            type = self._type, 
            exceptions = self._exceptions,
            on_before_addition = self.on_before_addition,
            on_after_addition = self.on_after_addition,
            on_before_deletion = self.on_before_deletion,
            on_after_deletion = self.on_after_deletion
        )
        result.extend(copy.deepcopy(self[:]))
        return result

    def __setitem__(self, key, value):
        self.__check(value)
        super(TypedList, self).__setitem__(key, value)
 
    def append(self, value):
        self.__check(value)
        self.__do_before_addition(value)
        super(TypedList, self).append(value)
        self.__do_after_addition(value)

    def insert(self, num, value):
        self.__check(value)
        self.__do_before_addition(value)
        super(TypedList, self).insert(num, value)
        self.__do_after_addition(value)
        
    def extend(self, values):
        map(self.append, values)

    def remove(self, value):
        if value in self:
            self.__do_before_deletion(value)
        super(TypedList, self).remove(value)
        self.__do_after_deletion(success=True)

    def pop(self, index=-1):
        if 0 <= index < len(self) or -len(self) <= index <= -1:
            self.__do_before_deletion(self[index])
        super(TypedList, self).pop(index)
        self.__do_after_deletion(success=True)

    def clear(self):
        '''
        Очищает список
        '''
        while len(self) > 0:
            self.pop()
        
    def __check(self, value):
        '''
            Функция проверки типа и значения
            Сначало проверяется значение, то есть если значение "Разделитель" = "-" и тд, то дальнейшие проверки не осуществляются
            Иначе проверяется тип объекта.
        '''
        if value not in self._exceptions:
            assert isinstance(value, self._type), 'Type value "%s" isn\'t %s!' % (value, self._type.__name__)

    def __do_before_addition(self, *arg, **kwargs):
        if self.__on_before_addition is not None:
            self.__on_before_addition(*arg, **kwargs)

    def __do_after_addition(self, *arg, **kwargs):
        if self.__on_after_addition is not None:
            self.__on_after_addition(*arg, **kwargs)

    def __do_before_deletion(self, *arg, **kwargs):
        if self.__on_before_deletion is not None:
            self.__on_before_deletion(*arg, **kwargs)

    def __do_after_deletion(self, *arg, **kwargs):
        if self.__on_after_deletion is not None:
            self.__on_after_deletion(*arg, **kwargs)

    @property
    def on_before_addition(self):
        return self.__on_before_addition

    @on_before_addition.setter
    def on_before_addition(self, handler):
        self.__on_before_addition = handler

    @property
    def on_after_addition(self):
        return self.__on_after_addition

    @on_after_addition.setter
    def on_after_addition(self, handler):
        self.__on_after_addition = handler

    @property
    def on_before_deletion(self):
        return self.__on_before_deletion

    @on_before_deletion.setter
    def on_before_deletion(self, handler):
        self.__on_before_deletion = handler

    @property
    def on_after_deletion(self):
        return self.__on_after_deletion

    @on_after_deletion.setter
    def on_after_deletion(self, handler):
        self.__on_after_deletion = handler


class MutableList(list):
    '''
    Список с признаком измененности "changed"
    '''
    def __init__(self, mutable = True, *args, **kwargs):
        super(MutableList, self).__init__(*args, **kwargs)
        self._mutable = mutable
        self.changed = False
    
    def _assert_mutable(self):
        if not self._mutable:
            raise AttributeError("This MutableList instance is immutable")
        self.changed = True
    
    def __setitem__(self, key, value):
        self._assert_mutable()
        super(MutableList, self).__setitem__(key, value)
        
    def append(self, value):
        self._assert_mutable()
        super(MutableList, self).append(value)
    
    def extend(self, values):
        self._assert_mutable()
        super(MutableList, self).extend(values)
    
    def insert(self, num, value):
        self._assert_mutable()
        super(MutableList, self).append(num, value)



#coding: utf-8
"""
Хелперы для удобного чтения конфигурационного файла проекта

@author: akvarats
"""

import ConfigParser

class ProjectConfig:
    """
    Обертка над парсером параметров конфигурации
    """
    def __init__(self, filenames=[], defaults={}):
        self.parser = ConfigParser.ConfigParser()
        if filenames:
            self.parser.read(filenames)
        self.defauls = defaults
        
    def read(self, filenames):
        """
        Загружает конфигурацию из файла(ов) filenames
        """
        self.parser = ConfigParser()
        self.parser.read(filenames)
    
    def set_defaults(self, defaults):
        """
        Устанавливает параметры проекта по умолчанию
        """
        self.defauls = defaults
        
    def items(self, section):
        """
        Возвращает список кортежей (key, value) из указанной секции.
        В случае, если такой секции нет, то возвращается пустой список.
        """
        return self.parser.items(section) if self.parser.has_section(section) else []
    
    def get(self, section, option):
        """
        Безопастно возвращает значение конфигурационного параметра option
        из раздела section
        """
        try:
            value = self.parser.get(section, option).strip()
            if not value:
                return self.defauls[(section,option)]
            return value
        except:
            if self.defauls.has_key((section,option)):
                return self.defauls[(section,option)]
        return ''
    
    def get_bool(self, section, option):
        """
        Безопастно возвращает булево из конфига
        """
        value = self.get(section, option)
        if isinstance(value, basestring):
            value = value.upper() == 'TRUE'

        return bool(value)

        
    def get_int(self, section, option):
        """
        Безопасно возвращает целое число из конфига
        """
        value = self.get(section, option)
        if isinstance(value, str):
            try:
                value = int(value)
            except:
                value = 0
        return value
        
    def get_uint(self, section, option):
        """
        Безопасно возвращает положительное целое число из конфига
        """
        value = self.get_int(section, option)
        if value < 0:
            value = 0
        return value
    


#coding:utf-8
"""
Created on 16.01.2011

@author: akvarats
"""
import datetime

try:
    from threading import local
except ImportError:
    from django.utils._threading_local import local

_thread_locals = local()

from django.conf import settings
from django.contrib import auth

try:
    from django.utils.log import logger
except ImportError:
    from django.utils.log import getLogger
    logger = getLogger('django')


#==============================================================================
# M3 Common Middleware
#==============================================================================

class ThreadData(object):
    """
    Класс-враппер для набора данных, которые храняться в
    thread-locals
    """

    DEFAULT_USER_ID = 0
    DEFAULT_USER_LOGIN = 'console-user'
    DEFAULT_USER_NAME = u'Пользователь системной консоли'

    DEFAULT_SESSION_KEY = 'console-session'

    DEFAULT_CLIENT_HOST = 'server'

    ANONYMOUS_USER_ID = 0
    ANONYMOUS_USER_LOGIN = 'anonymous-user'
    ANONYMOUS_USER_NAME = u'Анонимный пользователь'

    def __init__(self):
        self.clear()

    def clear(self):
        """
        Приводит данные объекта в начальное состояние
        """
        # user data
        self.user_id = None  # идентификатор пользователя
        self.user_login = ''  # Логин пользователя
        self.user_name = ''  # ФИО пользователя
        self._user_is_authenticated = False

        # session data
        self.session_key = ''  # идентификатор пользовательской сессии

        # client host data
        self.client_host = ''

    def apply_defaults(self):
        """
        Метод применяет дефолтные значения атрибутов
        """
        self._apply_user_defaults()
        self._apply_session_defaults()
        self._apply_client_host_defaults()

    def _apply_user_defaults(self):
        self.user_id = self.user_id or ThreadData.DEFAULT_USER_ID
        self.user_login = self.user_id or ThreadData.DEFAULT_USER_LOGIN
        self.user_name = self.user_id or ThreadData.DEFAULT_USER_NAME

    def _apply_session_defaults(self):
        self.session_key = self.user_id or ThreadData.DEFAULT_SESSION_KEY

    def _apply_client_host_defaults(self):
        self.client_host = self.client_host or ThreadData.DEFAULT_CLIENT_HOST

    def _apply_anonymous_user(self):
        """
        Проставляет значения, специфичные для анонимного пользователя системы
        """
        self.user_id = ThreadData.ANONYMOUS_USER_ID
        self.user_login = ThreadData.ANONYMOUS_USER_LOGIN
        self.user_name = ThreadData.ANONYMOUS_USER_NAME
        self._user_is_authenticated = False

    def read(self, request=None):
        """
        Читает данные из объекта request
        """
        self.clear()

        if request:
            # читаем данные пользователя
            user = auth.get_user(request)
            if user:
                self.user_is_authenticated = user.is_authenticated()
                if user.is_authenticated():
                    self.user_id = user.id
                    self.user_login = user.username
                    self.user_name = ('%s %s' % (user.first_name,
                                                 user.last_name)).strip()
                else:
                    self._apply_anonymous_user()
            else:
                self._apply_user_defaults()

            # читаем данные сессии
            if hasattr(request, 'session') and request.session:
                self.session_key = request.session.session_key
            else:
                self._apply_session_defaults()

            if hasattr(request, 'META') and request.META:
                self.client_host = request.META.get('HTTP_X_FORWARDED_FOR', '')
            else:
                self._apply_client_host_defaults()

        return self

    def dump(self):
        """
        Выводит состояние объекта в строку для получения отладочной информации
        """
        str = (
            '''session_key: %s
user_id: %s
user_login: %s
user_name: %s
client_host: %s'''
        ) % (
            self.session_key,
            self.user_id,
            self.user_login,
            self.user_name,
            self.client_host
        )
        return str


def get_thread_data():
    """
    Метод возвращает объект типа ThreadData, который описывает
    текущие thread-local параметры обработки запроса.
    """
    if not hasattr(_thread_locals, 'm3_data'):
        default_data = ThreadData()
        default_data.apply_defaults()
        _thread_locals.m3_data = default_data

    return _thread_locals.m3_data


class M3CommonMiddleware(object):
    """
    Middleware общего назначения для проектов на M3.

    Данное middleware должно располагаться ниже
    SessionMiddleware и AuthenticationMiddleware
    """
    def process_request(self, request):
        """
        Обработчик, срабатывающий перед выполнением запроса в прикладном
        приложении. Записываем в thread-locals информацию о
        текущей сессии и пользователе.
        """
        _thread_locals.m3_data = ThreadData()
        _thread_locals.m3_data.read(request)

        _thread_locals.foo = 'foo'


    def _clear(self):
        """
        Очищает информацию в thread-locals о текущей сессии и пользователе
        """
        if hasattr(_thread_locals, 'm3_data'):
            del _thread_locals.m3_data

    def process_response(self, request, response):
        self._clear()
        return response

    def process_exception(self, request, exception):
        self._clear()


class M3SimpleProfileMiddleware(object):
    """
    Выводит в отладочный лог время отработки задного запроса
    вместе с полным адресом
    """
    def process_request(self, request):
        request.m3_simple_profile_start_time = datetime.datetime.now()

    def process_response(self, request, response):
        try:
            logger.debug(
                str(
                    datetime.datetime.now() -
                    request.m3_simple_profile_start_time
                ) + ' ' + request.get_full_path())
        except:
            pass

        return response


class AutoLogout(object):
    """
    Отслеживает активность пользователей в системе.
    Если с последнего действия пользователя прошло времени
    больше чем INACTIVE_SESSION_LIFETIME,
    то он выводит пользователя из системы
    """

    session_key = 'app_last_user_activity_time'

    def process_request(self, request):
        # Если проверка отключена
        if settings.INACTIVE_SESSION_LIFETIME == 0:
            return

        # У аутентифицированного пользователя проверяем таймаут,
        # а ананимусов сразу посылаем
        if request.user.is_authenticated():
            last_time = request.session.get(self.session_key, None)
            if last_time is not None:
                delta = datetime.datetime.now() - last_time
                if delta.seconds / 60 > settings.INACTIVE_SESSION_LIFETIME:
                    # После логаута сессия уже другая
                    # и присваивать время не нужно
                    auth.logout(request)
                    return

            # Записываем время последнего запроса
            request.session[self.session_key] = datetime.datetime.now()
